# -*- coding: utf-8 -*-
from efr.utils import *
from efr.core.event import Event
from efr.core.equeue import EventQueue


class EventStation(EventQueue):
    def __init__(self, key:str=None, filter:callable=None, respond:callable=None, lv:int=0, step:int=None, timeout:int=None):
        """

        :param key:
        :param lv:
        :param step:
        """
        super(EventStation, self).__init__()
        self.efr = None

        self.key = key if key else self.__class__.__name__
        self.lv = lv

        self._worker = None
        self._task = None
        self._filter = filter
        self._respond = respond

        self.step = step
        self.timeout = timeout

    # @abstractmethod
    def filter(self, event:Event) -> bool:
        """
        如果初始化时未传入filter则需要用户覆盖此方法
        If filter is not passed in during initialization, you need to override this method

        定义工作站对事件的筛选方法. 筛选通过返回True, 失败返回False
        Defines how the workstation filters events. The filter returns true if passed and false if failed

        默认筛选方法是: 判断事件的dest是否和工作站的key匹配
        The default filtering method is to judge whether the dest of the event matches the workstation.key

        :param event: Event
        :return: bool
        """
        if self._filter:
            return self._filter(event)
        return self.key == event.dest

    # @abstractmethod
    def respond(self, event:Event) -> object:
        """
        如果初始化时未传入filter则需要用户覆盖此方法
        If filter is not passed in during initialization, you need to override this method

        :param event: Event
        :return: object
        """
        if self._respond:
            return self._respond(event)
        return

    def __eq__(self, other):
        return isinstance(other, EventStation) and self.key == other.key and self._filter == other._filter and self._respond == other._respond \
               and self.filter == other.filter and self.respond == other.respond

    def push(self, event:Event, timeout=1) -> bool:
        """

        :param event:
        :return: bool
        """
        event = super(EventStation, self).push(event, timeout)
        if event:
            event.result[self.key] = None
            if event.state is Event.STATE_JUNIOR:
                event.state = Event.STATE_URGENT
            event.links += 1
            return True
        return False

    def update(self):
        events = self.release(self.step, self.timeout)
        for event in events:
            if event.state is not Event.STATE_EXCEPT:
                try:
                    ret = self.respond(event)
                except Exception as err:
                    event.state = Event.STATE_EXCEPT
                    err.trace = traceback.format_exc()
                    event.trace.update(err, self.efr)

                if event.state is not Event.STATE_EXCEPT:
                    event.setResult(self.key, ret, syscall=True)
                    event.links -= 1
                    if event.links <= 0:
                        event.state = Event.STATE_RETIRED
                    else:
                        event.state = Event.STATE_FINISH




def singleton(cls, *args, **kwargs):
    instances = {}

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return _singleton

def EmptyFunction(*args, **kwargs):
    ...



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
import random
import time

from efr import EventFramework, EventStation, Event, Task

class ExceptTest(EventStation):
    def filter(self, event:Event) -> bool:
        if event.dest == 'except':
            if event.task % 3 == 0:
                raise Exception("Error hanpen in filter")
            return True
    def respond(self, event) -> object:
        if event.task % 3 == 1:
            raise Exception("Error hanpen in respond")
        return "正常执行"

"""
主线程承担Producer，负责产生event
在事件框架中测试异常处理

事件框架执行中如果出现错误，并不会直接报错，而是由对应的event记录

"""

efr = EventFramework(log_path='test.log')

except_test = ExceptTest( 'err' )  # 新建工作站
efr.login(except_test)  # 注册到事件框架
efr.assign(efr.default_worker, except_test)  # 为test提供一个Worker用于更新

efr.start()  # 启动事件框架

events = []
for i in range(10):
    # 发送事件
    event = efr.push( Event(source="main", dest='except', task=i) )  # 推动事件进入事件循环
    events += [event]

    # join;
    time.sleep(0.5)

    # 读取返回值、更新events
    temp = []
    for event in events:
        if event.isFinish():  # 判断事件是否结束
            print( event.getResult('err') )
        elif event.isExcept():
            print("event except.", event.task)
            for err in event.getExcepts():
                print(event.trace.strError(err))
                # 可以在程序和日志中看到错误信息
                # 隐藏错误: SolutionMissing:
                #       当filter错误时，会继续寻找其他station去filter这个事件
                #       当没有任何station接受此event时抛给event错误:SolutionMissing. 然后抛弃该event
            print()
        else:
            temp += [event]
    events = temp


# 退出框架
efr.quit()



# -*- coding: utf-8 -*-
class SolutionMissing(Exception):
    def __init__(self, *args, event=None):
        super(SolutionMissing, self).__init__(*args)
        self.event = event

    def __str__(self):
        txt = '\n[SolutionMissing]: ' + "Event refused by all stations.\n"
        if self.event:
            txt += "\tEvent: " + str(self.event)
        if self.args:
            txt += "\tDetails:"
            for info in self.args:
                txt += "\n\t\t" + str(info)
        return txt



if __name__ == '__main__':
    raise SolutionMissing()



import os
from efr.utils import *
class Trace:
    """
    .caller: Trace的直接调用者
    一般而言，Trace的调用者才是我们关心的对象，暂称其为target
    .initial: target的调用者
    """
    def __init__(self):
        stack:list = traceback.extract_stack()
        self.caller  = stack[-2][2]
        self.initial = stack[-3][2]
        self._errs = []

    @property
    def reason(self):
        return self._errs

    def update(self, err:Exception=None, efr:object=None):
        """
        Update trace.
        :param err: Exception
        :param efr: EventFramework
        :return:
        """
        if err:
            self._errs += [err]
        if efr.log_immediate:
            efr._logger.error(self.strError(err))
        else:
            efr.quit_logs += [err]

    @staticmethod
    def strError(error):
        def _inner_recur(fr, stack):
            if fr.f_back:
                _inner_recur(fr.f_back, stack)

            _sfr = str(fr)
            index = _sfr.find('file')
            if index != -1:
                _sfr = _sfr[index:-1]
            stack += [_sfr]
        if isinstance(error, Exception):
            tr = getattr(error, 'trace', None)
            if tr:
                return tr  # + "\n\t" + str(error)
            tb = error.__traceback__
            if tb:
                fr = tb.tb_frame
                if fr:
                    stack = []
                    _inner_recur(fr, stack)

                    txt = "Exception Traceback: \n"
                    for i in range(len(stack)):
                        str_frame = stack[i]
                        txt += '  ' * (i + 1) + str_frame + "\n"
                    base = "  " * ( len(stack) + 1 )
                    txt += base + ">{} At line:{}:".format(error.__class__.__name__, fr.f_lineno) + '\n'

                    _lis  = str(error).split("\n")
                    for _line in _lis:
                        if _line:
                            txt += base + _line + '\n'

                    return txt

        return str(error)


if __name__ == '__main__':
    def target():
        tr = Trace()
        print(tr)
        print(tr.reason)

    def caller():
        target()

    caller()



# -*- coding: utf-8 -*-
import traceback

from efr.core.event import Event
from efr.core.equeue import EventQueue
from efr.core.estation import EventStation

from efr.core.eerrors import SolutionMissing


class EventAlloter:
    """
    事件分配器
    事件分配器会在update函数被调用时工作，取出事件队列中的事件，然后轮询所有已注册的工作站(根据事件的性质可能会提前结束). 最后将事件分配到对应的工作站的队列中

    Event Allocator
    The event allocator will work when the 'update' function is called, take out the events in the event queue, and then poll all registered workstations (depending on the nature of the event, it may end in advance) Finally, the event is allocated to the queue of the corresponding workstation
    """
    def __init__(self, equeue:EventQueue, step:int=None, timeout:int=None):
        """

        :param equeue:
        :param step: int Update amount per update
                _: None  # mean inf
        :param timeout:float .Wait for more than this time when update
                _:None  # mean inf
        """
        self.efr = None

        self.equeue = equeue
        self.step = step
        self.timout = timeout

        self.stations = []

    def login(self, station:EventStation)->bool:
        """
        注册事件工作站。当update时，alloter会用调用station.filter(事件)->bool来判断station是否响应此事件。如果station响应了事件，那么alloter会把事件传入该station的队列.
        成功返回True，失败返回False

        Register the event workstation. When updating, the alloter will call station Filter - > bool to determine whether the station responds to this event. If the station responds to an event, the alloter will pass the event into the queue of the station
        Returns true for success and false for failure
        :param station:
        :return: bool
        """
        if station in self.stations:
            return False

        station.efr = self.efr

        for i in range(len(self.stations)):
            if station.lv > self.stations[i].lv:
                self.stations.insert(i, station)
                return True
        self.stations += [station]
        return True

    def logoff(self, station:EventStation)->bool:
        """
        注销工作站。成功或不存在station返回True，失败返回False
        Log off the eventstation. Successful or non-existent stations return true, while failure returns false
        :param station:
        :return: bool
        """
        try:
            station.etr = None
            self.stations.remove(station)
        except:
            ...

        return True

    def update(self):
        """
        由外部单元执行更新
        Updates are performed by external units
        :return:
        """
        events = self.equeue.release(self.step, self.timout)

        for event in events:
            flag = False
            for station in self.stations:
                # try filter.
                try:
                    get = station.filter(event)
                except Exception as err:
                    event.state = Event.STATE_EXCEPT
                    err.trace = traceback.format_exc()
                    event.trace.update(err, self.efr)
                    continue

                # add to station
                if get:
                    get = station.push(event)
                    if get:
                        flag = True

                    event.times -= 1
                    if event.times <= 0:
                        break

            # set exception
            if not flag:
                event.state = Event.STATE_EXCEPT
                # record except
                try:
                    raise SolutionMissing(event)
                except SolutionMissing as err:
                    err.trace = traceback.format_exc()
                    event.trace.update(err, self.efr)



# -*- coding: utf-8 -*-
from efr.core.event import Event
from efr.core.equeue import EventQueue
from efr.core.ealloter import EventAlloter
from efr.core.estation import EventStation
from efr.core.eframework import EventFramework




# -*- coding: utf-8 -*-
import random
import time

from efr import EventFramework, EventStation, Event, Task

def respond(event) -> object:
    a, b = event.task
    return a + b

"""
主线程承担Producer，负责产生event
在事件框架中添加一个Comsumer负责消耗event
    comsumer负责进行加法，然后返回
"""

efr = EventFramework(log_path='test.log')

# 添加comsumer
comsumer = EventStation( 'comsumer', respond = respond )  # 新建工作站
efr.login(comsumer)  # 注册到事件框架
efr.assign(efr.default_worker, comsumer)  # 为comsumer提供一个Worker用于更新

efr.start()  # 启动事件框架

events = []
for i in range(10):
    # 发送事件
    a, b = random.randint(0, 50), random.randint(0, 50)
    event = efr.push( Event(source="main", dest='comsumer', task=[a, b]) )  # 推动事件进入事件循环
    events += [event]

    # join;
    time.sleep(0.5)

    # 读取返回值、更新events
    temp = []
    for event in events:
        if event.isFinish():  # 判断事件是否结束
            print( '{:^4d} + {:^4d} = {:^4d}'.format(*event.task, event.getResult('comsumer')) )
        else:
            temp += [event]
    events = temp

# 退出框架
efr.quit()



from random import randint

# 10**n 查找表
searchMax = 32
searchTable = []
for i in range(searchMax):
    searchTable.append(10*10**i - 1)

def NewRandom(ulen=16):  # 本机测试:1.5us生成1次
    assert 0 < ulen <= searchMax, "[ulen] can not over than {} and less than 0".format(searchMax)
    return randint(0, searchTable[ulen-1])


def NewRandomID(ulen=16):  # 本机测试:3us生成1次
    return ("{:0>" + str(ulen) + "d}").format(NewRandom(ulen))


if __name__ == '__main__':
    @fn_timer
    def test(times, ulen=32):
        for i in range(times):
            NewRandomID(ulen)

    print(NewRandomID(16))
    test(10000)
    test(10000)
    test(10000)
    test(10000)


# -*- coding: utf-8 -*-
from efr.utils import inf
from efr.utils.Functions import EmptyFunction

ONCE = 1
CIRCLE = inf

class Task:
    ONCE = ONCE
    CIRCLE = CIRCLE
    def __init__(self, target=EmptyFunction, args=(), kwargs={}, times=ONCE):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.left = times

    def __call__(self, *args, **kwargs):
        self.left -= 1
        return self.target(*args, *self.args, **kwargs, **self.kwargs)

if __name__ == '__main__':
    task = Task(lambda a: print(a))
    task("hello, world!")



# -*- coding: utf-8 -*-
from efr.utils import *

class EventInterface(ABC):
    @property
    @abstractmethod
    # 事件追踪器
    def trace(self): ...

    @property
    @abstractmethod
    # 事件任务
    def task(self): ...

    @property
    @abstractmethod
    # 事件的task的future，这里进行了简化
    def result(self): ...

    @property
    @abstractmethod
    # 事件来源
    def source(self): ...

    @property
    @abstractmethod
    # 事件去向
    def dest(self): ...

    @property
    @abstractmethod
    # 事件具有的tag
    def tags(self): ...

    @property
    @abstractmethod
    # 事件状态
    def state(self): ...

    @property
    @abstractmethod
    # 事件连接数
    def links(self): ...

class EventState:
    def __str__(self):
        return self.__class__.__name__

@singleton
class State_OffLine(EventState): ...
@singleton
class State_Junior (EventState): ...
@singleton
class State_Urgent (EventState): ...
@singleton
class State_Finish (EventState): ...
@singleton
class State_Retired(EventState): ...
@singleton
class State_Except (EventState): ...

STATE_OFFLINE = State_OffLine()
STATE_JUNIOR  = State_Junior()
STATE_URGENT  = State_Urgent()
STATE_FINISH  = State_Finish()
STATE_RETIRED = State_Retired()
STATE_EXCEPT  = State_Except()

EVENT_ONCE = 1
EVENT_ALL = inf

class Event:
    # static const
    STATE_OFFLINE = STATE_OFFLINE
    STATE_JUNIOR  = STATE_JUNIOR
    STATE_URGENT  = STATE_URGENT
    STATE_FINISH  = STATE_FINISH
    STATE_RETIRED = STATE_RETIRED
    STATE_EXCEPT  = STATE_EXCEPT

    EVENT_ONCE = EVENT_ONCE
    EVENT_ALL  = EVENT_ALL

    """
    事件通常有5个状态：
    1.STATE_OFFLINE - 游离态
        刚刚实例化的事件对象，尚未添加到事件框架中。
    2.STATE_JUNIOR - 预备态
        该事件被EventFramework.push()到了事件队列。但尚未被进一步处理的状态
    3.STATE_URGENT - 激发态
        该事件已经被分配到了EventStation中，但尚未被执行的状态
    4.STATE_FINISH - 完成态
        该事件至少被成功执行了一次
        此时的Event可以通过Event.result获取部分结果
    5.STATE_RETIRED - 末态
        该事件被所有工作站成功执行, 并且已经脱离了EventFramework
        此时的Event可以通过Event.result获取所有结果
    *6.STATE_EXCEPT - 异常态
        该事件在任意环节中出错，通过Event.trace获取更多错误信息

    The event generally has four states:
    1.STATE_OFFLINE - free state
        The event object just instantiated has not been added to the EventFramework.
    2.STATE_JUNIOR - ready
        The event was by EventFramework.push() to the event queue. A state that has not yet been further processed
    3.STATE_URGENT - excited state
        The state in which the event has been assigned to an EventStation but has not yet been executed
    4.STATE_FINISH - finish atleast once
        The event was executed successfully atleast once and still in EventFramework
        At this time, the event can be through Event.result get the part of the results
    5.STATE_RETIRED - final state
        The event was executed successfully in all task and has broken away from the EventFramework
        At this time, the event can be through Event.result get the all the results
    *6.STATE_EXCEPT - exception
        The event has an error in any step, which is passed through Event.trace for more error information
    """
    def __init__(self, source:Any=None, dest:Any=None, task:Any=None, tags=(), times:int=EVENT_ALL):
        """

        :param source:
        :param dest:
        :param task:
        :param tags:
        :param times:
        """
        self.efr = None

        self.trace = Trace()
        self.task = task
        self.result = {}
        self.source = source
        self.dest = dest
        self.tags = tags
        self.state = STATE_OFFLINE
        self.times = times
        self.links:int = 0

    def isOffline(self) -> bool:
        """
        检查事件是否处于游离状态
        Check whether the event is offline
        :return: bool
        """
        return self.state is STATE_OFFLINE

    def isJunior(self) -> bool:
        """
        检查事件是否已经被添加到事件队列中或已经处于urgent状态
        Check whether the event has been added to eventQueue
        :return: bool
        """
        return self.state is STATE_JUNIOR or self.state is STATE_URGENT

    def isUrgent(self) -> bool:
        """
        检查事件是否已经被添加到station中
        Check whether the event has been added to station
        :return: bool
        """
        return self.state is STATE_URGENT

    def isFinish(self) -> bool:
        """
        检查事件是否已经结束或完全结束。只有结束的事件才可以获取result
        Check whether the event has finished or retired. Only the finished event can get the result.
        :return: bool
        """
        return self.state is STATE_FINISH or self.state is STATE_RETIRED

    def isRetired(self) -> bool:
        """
        检查事件是否已经完全结束。
        Check whether the event has retired.
        :return: bool
        """
        return self.state is STATE_RETIRED

    def isExcept(self) -> bool:
        """
        检查事件是否出现错误
        Check whether the event has except.
        :return: bool
        """
        return self.state is STATE_EXCEPT

    def getExcepts(self) -> list:
        """
        获取event的错误信息. 尚未出错或失败返回[]
        Get the exception-information of event. No error or failure returned []
        :return: list of Exceptions
        """
        return self.trace.reason

    def waitFor(self, state:EventState=STATE_RETIRED, timeout:float=None, timedt:float=0.1) -> bool:
        """
        等待事件被执行到某种状态。成功返回True，失败返回False
        Wait for the event to be executed to a certain state. Returns true for success and false for failure
        :param state  : EventState, Used to indicate the waiting target
                _: STATE_RETIRED
        :param timeout: float, Wait for more than this time to force the end and return false
                _: None  # mean inf
        :param timedt : float, While waiting, the result is checked every time the 'timedt' elapses(unit: s)
                _: 0.1
        :return: bool
        """
        if timeout:
            end_point:float = timeout if timeout else inf
        while self.state is not state:
            if timeout and time.time() > end_point:
                return False
            time.sleep(timedt)
        return True

    def setResult(self, station_key:object, result:object, syscall=False):
        """
        set the result of event
        (API for outer caller? maybe)
        :param station_key: object
        :param result: object from respond
        :param syscall: bool Whether call from eventframework. If not, it whill auto set it's state.
        :return:
        """
        self.result[station_key] = result
        if not syscall:
            self.links -= 1
            if self.links <= 0:
                self.state = STATE_FINISH
            else:
                self.state = STATE_RETIRED

    def getResult(self, station_key: object)->object:
        """
        get the result of event from station
        :param station_key: object
        :return: object
        """
        return self.result.get(station_key)

    # def record(self):
    #     """
    #     提供该event的当前的路径(用于多线程trace)，推荐用户不操作该函数
    #     Provide the current trace of the event (for multi-threaded trace). It is recommended that users do not operate this function
    #     :return:
    #     """
    #     self.trace.update(traceback.extract_stack())

    def __str__(self):
        txt = "Event[source:{}, dest:{}, task:{}, tags:{}]".format(self.source, self.dest, self.task, self.tags)
        return txt

if __name__ == '__main__':
    s2 = State_OffLine()
    print(s2 is STATE_OFFLINE)




#!/usr/bin/env python
# coding:utf-8

from setuptools import find_packages, setup

setup(
name='efr',
version='0.1.8',
description='Event-Driver (Stand-alone) Framework. Pure python. Efficient.',
author="Eagle'sBaby",
author_email='2229066748@qq.com',
maintainer="Eagle'sBaby",
maintainer_email='2229066748@qq.com',
packages=find_packages(),
platforms=["all"],
license='Apache Licence 2.0',
classifiers=[
'Programming Language :: Python',
'Programming Language :: Python :: 3',
],
keywords = ["event-drive"],
python_requires='>=3', 
)


# -*- coding: utf-8 -*-
from efr.utils import *
from efr.core.event import Event
"""
事件队列
"""

class EventQueue:
    """
    事件队列，事件框架的事件入口

    EventQueue, event entry of EventFramework
    """
    def __init__(self, capacity=None):
        """

        :param capacity: capacity of QventQueue
                _: None  # mean inf
        """
        self.efr = None
        self._capacity = capacity if capacity else inf
        self.equeue = Queue(0 if capacity is None else capacity)

    def push(self, event:Event, timeout:float=None) -> Event:
        """
        驱动事件。成功返回该event, 失败返回None
        :param event: Event Event instance.
        :param timeout:float .Wait for more than this time to force the end and return None (if queue refused.)
                _:None  # mean inf
        :return: Event
        """
        get = self.equeue.put(event, block=True, timeout=timeout)
        event.efr = self.efr
        if get is not QFull:
            if event.state is Event.STATE_OFFLINE:
                event.state = Event.STATE_JUNIOR
            return event
        return None

    def getSize(self):
        """
        Get how many event in queue.
        equal to len(self)
        :return: int
        """
        return self.equeue.qsize()

    def getMaxSize(self):
        """
        Get the capacity of EventQueue
        :return:
        """
        return self._capacity

    def release(self, num:int=None, timeout:float=None):
        """
        获取事件队列中的事件, 该函数不推荐用户使用。用户调用会导致EventFramework工作异常
        Get events in the event queue. This function is not recommended for users. User calls will cause the EventFramework to work abnormally
        :param num: int Get the number of events at one time. The number of returned events will not exceed this value.
                _: None  # mean inf
        :param timeout:float .Wait for more than this time to force the end and return list early (if queue refused.)
                _:None  # mean inf
        :return:
        """
        if num is None:
            num = inf

        _len:int = len(self)
        num = _len if _len < num else num
        ret: list = []
        while num:
            get = self.equeue.get(block=True, timeout=timeout)
            if get is QEmpty:
                return ret
            ret += [get]
            num -= 1
        return ret

    def __len__(self):
        return self.equeue.qsize()


# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
import random
import time

from efr import EventFramework, EventStation, Event, Task

class Calculator(EventStation):
    def filter(self, event) -> bool:
        return event.dest == 'calc'

class Adder(Calculator):
    def respond(self, event) -> object:
        a, b = event.task
        return a + b

class Subber(Calculator):
    def respond(self, event) -> object:
        a, b = event.task
        return a - b

class Multipler(Calculator):
    def respond(self, event) -> object:
        a, b = event.task
        return a * b

class Divver(Calculator):
    def respond(self, event) -> object:
        a, b = event.task
        return a / b

"""
主线程承担Producer，负责产生event
在事件框架中添加+-*/, 他们均对calc事件敏感
"""

efr = EventFramework(log_path='test.log')

# 添加stations
calcs = [Adder('+'), Subber('-'), Multipler('*'), Divver('/')]
# calcs = [Adder('+'), Subber('-'), Multipler('*'), Divver('/', lv = 1)]  # 指定lv可以改变顺序(默认都是0, 越大越优先)
worker = efr.addWorker()  # 新建一个worker用于处理这四个station的工作任务
for station in calcs:
    efr.login(station)  # 注册到事件框架
    efr.assign(worker, station)  # 为这些station提供一个Worker用于更新

efr.start()  # 启动事件框架

events = []
for i in range(10):
    # 发送事件
    a, b = random.randint(0, 50), random.randint(0, 50)
    event = efr.push( Event(source="main", dest='calc', task=[a, b]) )  # 推动事件进入事件循环(被多个事件处理)
    # event = efr.push( Event(source="main", dest='calc', task=[a, b], times=1) )  # 指定times可以限制event被处理的次数

    events += [event]

    # join;
    time.sleep(0.5)

    # 读取返回值、更新events
    temp = []
    for event in events:
        if event.isRetired():  # 判断事件是否完全结束
            print( '{:^4} + {:^4} = {:^4}'.format(*event.task, event.getResult('+')) )
            print( '{:^4} - {:^4} = {:^4}'.format(*event.task, event.getResult('-')) )
            print( '{:^4} * {:^4} = {:^4}'.format(*event.task, event.getResult('*')) )
            print( '{:^4} / {:^4} = {:^4}'.format(*event.task, event.getResult('/')) )
            print()
        else:
            temp += [event]
    events = temp


# 退出框架
efr.quit()




import time
import logging
import inspect
import traceback
from math import inf
from typing import Any
from queue import Queue, Empty as QEmpty, Full as QFull
from abc import ABC, abstractmethod


# inner
from efr.utils.IdGenerator import NewRandomID, NewRandom
from efr.utils.Functions import EmptyFunction, singleton
from efr.utils.Trace import Trace
from efr.utils.Worker import Worker
from efr.utils.Task import Task

if __name__ == '__main__':
    print(inf > inf)
    print(inf < inf)



# -*- coding: utf-8 -*-
from efr.utils import *
from efr.core.event import Event
from efr.core.equeue import EventQueue
from efr.core.ealloter import EventAlloter
from efr.core.estation import EventStation


class _PARAM_NONE: ...


class _EventFramework:
    def __init__(self):
        self._worker_id = 0
        self._start_flag = False
        self._log_handler = None

    def _newworker(self, timedt=None):
        self._worker_id += 1
        worker = Worker("worker" + str(self._worker_id - 1) + " of " + self.name, mindt= timedt if timedt else self.timedt)
        if self._start_flag:
            worker.start()
        return worker

    def _initlogger(self):
        self._logger.setLevel(level=self.log_level)
        if self._log_handler:
            self._logger.removeHandler(self._log_handler)
        if self.log_path:
            open(self.log_path, 'w').close()
            self._log_handler = logging.FileHandler(self.log_path)
            self._log_handler.setLevel(level=self.log_level)
            formatter = logging.Formatter(self.log_format)
            self._log_handler.setFormatter(formatter)
            self._logger.addHandler(self._log_handler)

        return self._logger

    def _outputlogs(self):
        for log in self.quit_logs:
            if isinstance(log, Exception):
                self._logger.error(Trace.strError(log))
            elif isinstance(log, Warning):
                self._logger.warn(Trace.strError(log))
            else:
                self._logger.info(log)

    def _log(self, item):
        if self.log_path:
            self._logger.info(item)
        else:
            self.quit_logs += [item]


class EventFramework(_EventFramework):
    def __init__(self, name: str = None, capacity: int = None, step: int = None, timeout=None, timedt=0.1,
                 log_path=None, log_level=logging.INFO, log_immediate=True,
                 log_format='[%(asctime)s][%(levelname)s] - %(message)s'):
        """
        高效事件驱动框架
        Efficient event driven framework

        :param name: str EventFrame's name.
                _: None  # use default name
        :param capacity: int EventQueue's capacity. Mean the max size of eventqueue.
                _: None  # infinite queue capacity.
        :param step: int EventAlloter's step. Mean how many events the eventalloter allote per update.
                _: None
        :param timeout: int EventAlloter's timeout. Mean how long eventalloter wait for eventqueue
                _: None
        :param timedt: int Worker's timedt. Mean how frequency the Worker work
                _: 0.1  # worker do tasks per 0.1s in the quickest case
        :param log_path: str log path.
                _: None  # do not log.
        :param log_level: int log level.
                _: logging.INFO
        :param log_immediate: bool Is log immediatelly written.
                _: False  # only log when .quit()
        :param log_format: str log format.
                _: '[%(asctime)s][%(levelname)s] - %(message)s'
        """
        super(EventFramework, self).__init__()

        self.name = ("efr-" + name) if name else ("efr-" + NewRandomID(8))
        self.timedt = timedt
        self.log_path = log_path
        self.log_level = log_level
        self.log_format = log_format
        self.log_immediate = log_immediate
        self.quit_logs = []  # quit时一起处理

        # workers
        self.workers = []
        self.default_worker = self.addWorker()
        self.workers[0].addTask(Task(self.update, times=Task.CIRCLE))

        # 外部组件
        self.equeue = EventQueue(capacity)
        self.ealloter = EventAlloter(self.equeue, step, timeout)
        self._logger = logging.getLogger(self.name)

        # 启动event组件
        self.equeue.efr = self
        self.ealloter.efr = self
        self._initlogger()

    def start(self):
        """
        启动事件框架
        start eventframe
        :return:
        """
        self._log(self.name + "start.")
        self._start_flag = True
        for worker in self.workers:
            worker.start()

    def quit(self):
        """
        退出事件框架
        quit eventframework
        :return:
        """
        for worker in self.workers:
            worker.stop()
        self._outputlogs()
        self._log(self.name + "quit.")

    def login(self, station: EventStation, worker: Worker=None) -> bool:
        """
        注册事件工作站。当update时，alloter会用调用station.filter(事件)->bool来判断station是否响应此事件。如果station响应了事件，那么alloter会把事件传入该station的队列.
        成功返回True，失败返回False

        Register the event workstation. When updating, the alloter will call station Filter - > bool to determine whether the station responds to this event. If the station responds to an event, the alloter will pass the event into the queue of the station
        Returns true for success and false for failure
        :param station: EventStation pass through Event Station to login it
        :param worker: Worker assign worker to this station
                _: None  # do not assign.
        :return: bool
        """
        ret = self.ealloter.login(station)
        if ret and worker:
            self.assign(worker, station)
            # print(station._worker)
        return ret

    def logoff(self, station: EventStation) -> bool:
        """
        注销工作站。成功或不存在station返回True，失败返回False
        Log off the eventstation. Successful or non-existent stations return true, while failure returns false
        :param station:
        :return: bool
        """
        ret = self.ealloter.logoff(station)
        if ret:
            self.assign(None, station)
        return ret



    def push(self, event: Event, timeout: float = None) -> Event:
        """
        驱动事件。成功返回该event, 失败返回None
        :param event: Event Event instance.
        :param timeout:float .Wait for more than this time to force the end and return None (if queue refused.)
                _:None  # mean inf
        :return: Event
        """
        return self.equeue.push(event, timeout)

    def config(self, step: int = _PARAM_NONE, timeout: int = _PARAM_NONE, log_path: str = _PARAM_NONE,
               log_immediate: bool = _PARAM_NONE, log_format: str = _PARAM_NONE, log_level: int = _PARAM_NONE):
        """
        配置事件框架
        config it.
        :param step: int EventAlloter's step. Mean how many events the eventalloter allote per update.
                _: None
        :param timeout: int EventAlloter's timeout. Mean how long eventalloter wait for eventqueue
                _: None
        :param timedt: int Worker's timedt. Mean how frequency the Worker work
                _: 0.1  # worker do tasks per 0.1s in the quickest case
        :param log_path: str log path.
                _: None  # do not log.
        :param log_level: int log level.
                _: logging.INFO
        :param log_immediate: bool Is log immediatelly written.
                _: False  # only log when .quit()
        :param log_format: str log format.
                _: '[%(asctime)s][%(levelname)s] - %(message)s'
        :return:
        """

        if step is not _PARAM_NONE:
            self.ealloter.step = step
        if timeout is not _PARAM_NONE:
            self.ealloter.timeout = timeout
        if log_path is not _PARAM_NONE:
            self.log_path = log_path
            self._initlogger()
        if log_immediate is not _PARAM_NONE:
            if self.log_immediate == False and log_immediate == True:
                self._outputlogs()
            self.log_immediate = log_immediate
        if log_format is not _PARAM_NONE:
            self.log_format = log_format
            self._initlogger()
        if log_level is not _PARAM_NONE:
            self.log_level = log_level
            self._initlogger()

    def assign(self, worker: Worker, station: EventStation) -> bool:
        """
        布置任务到worker
        assign task to worker
        :param worker: Worker  If worker is None, mean cancel the station's worker
        :param station: Station
        :return: bool
        """
        # print(worker, station)
        if worker is None :
            if station._worker:
                ret = station._worker.removeTask(station._task)
                station._worker = worker
                station._task = None
                return ret
            return True
        else:
            if station not in self.ealloter.stations:
                self.login(station)
            station._worker = worker
            station._task = Task(station.update, times=Task.CIRCLE)
            return worker.addTask(station._task)


    def addWorker(self, timedt:int=None) -> Worker:
        """
        添加一个工人到事件框架
        add a worker to eventframework.
        :param timedt: int Worker's timedt. Mean how frequency the Worker work
                _: None  # mean use the default timedt defined in EventFramework.__init__
        :return: Worker
        """
        worker = self._newworker(timedt)
        self.workers += [worker]
        return worker

    def removeWorker(self, worker: Worker) -> bool:
        """
        移除指定的工人，对应的工作也会尽可能快地被搁置
        Remove the designated workers and the corresponding work will be put on hold as soon as possible
        :param worker: Worker
        :return: bool
        """
        try:
            self.workers.remove(worker)
            worker.stop()
        except ValueError:
            ...
        return True

    def update(self):
        """
        事件框架更新, 自动添加到worker 0
        efr update， auto added to worker 0
        :return: None
        """
        return self.ealloter.update()


if __name__ == '__main__':
    none = _PARAM_NONE()
    print(none)



import time
from threading import Thread
from efr.utils.Task import *

class Worker(Thread):
    """
    提供一个循环执行器线程
    """
    def __init__(self, name=None, daemon=True, mindt=0.5):
        super(Worker, self).__init__(name=name, daemon=daemon)
        self.alive = True
        self.mindt = mindt
        self.tasks = []

    def run(self) -> None:
        while self.alive:
            next_point = time.time() + self.mindt

            tasks = []
            for task in self.tasks:
                task()
                if task.left > 0:
                    tasks += [task]
                if not self.alive:
                    break
            self.tasks = tasks

            delta = next_point - time.time()
            if delta > 0:
                time.sleep(delta)


    def addTask(self, task:Task) -> bool:
        self.tasks += [task]
        return True

    def removeTask(self, task:Task) -> bool:
        """
        移除任务
        remove task
        :param task: Task
        :return:
        """
        try:
            self.tasks.remove(task)
        except:
            ...
        return True

    def stop(self):
        """
        停止线程，但是至少需要等待到进行中的任务结束
        Stop the thread, but at least wait until the end of the task in progress
        :return:
        """
        self.alive = False
